/*
 * *    source version @(#) fanet.c 1.8@(#)
 *
 * "based" on my pci "base 'driver, from http://www.bolthole.com/solaris/
 *
 * This particular source is licensed under LGPL
 */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>

#include <sys/stropts.h>
#include <sys/stream.h>
#include <sys/dlpi.h>
#include <sys/gld.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [fanet_cb_ops, fanet_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	fanet_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	fanet_probe(dev_info_t *dip);
static	int	fanet_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);

static	int	fanet_print(dev_t, char *);

static	int	fanet_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

static	int	fanet_open(dev_t *devp, int flag, int otyp,cred_t *cred_p);
static	int	fanet_close(dev_t dev, int flag, int otyp,cred_t *cred_p);


char _depends_on[] = "misc/gld";

/* "char/block operations" */
static struct cb_ops	fanet_cb_ops = {
	fanet_open,		/* open()   */
	fanet_close,		/* close()  */
	nodev,			/* strategy()		*/
	fanet_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	nodev,		/* read() */
	nodev,		/* write() */
	nodev,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
	0,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops fanet_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	gld_getinfo,			/* devo_getinfo done by gld*/
	nulldev,		/* devo_identify */
	fanet_probe,		/* devo_probe */
	fanet_attach,		/* devo_attach */
	fanet_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&fanet_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"Experimental _fanet_ driver",
	&fanet_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/

/* This describes things like byteorder internally for the ethernet card */
static struct ddi_device_acc_attr dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};

#define BASE_MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   fanet_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * fanet_glob_soft_handle;


struct fanet_softstate {
	dev_info_t *dip; /* you'll probably want this */
	kmutex_t instance_mutex, intr_mutex,HIintr_mutex;
	gld_mac_info_t *gld_info;
	
	ddi_acc_handle_t fanet_regs_handle; /*opaque kernel stuff to map regs*/
	struct fanet_regs *fanet_regs; /* actual pointer to mapped registerset*/

	/******************************************/

	ddi_iblock_cookie_t intr_cookie, HIintr_cookie;

	ddi_softintr_t softintr_id;
	int soft_trigger; /* number of times triggered, vs serviced */

	int intr_hardcount,intr_softcount; /* debug info */


	uint8_t use_softintr; /* flag if we get stuck with highlevel intrs*/

	/******************************************/
	/* the statistics section */

	long long rbytes64;
	uint32_t rbytes;
	long long obytes64;
	uint32_t obytes;
	long long ipackets64;
	uint32_t ipackets;
	long long opackets64;
	uint32_t opackets;
	long multirvc;	/* multicast recieved */
	long multixmt;	/* multicast transmitted */
	long brdcstrvc;
	long brdcstxmt;
	long unknowns;
	long noxmitbuf; /* bufferoverflow/busy */
	long blocked;
	long xmtretry;
	int promisc;	/* am promiscuous? */

	long long ifspeed;
	int media; /*?? */
	int intr; /* interrupt count */
	int norcvbuf;
	int ierrors, oerrors;
	int missed; /* botched RECIEVED packets */
	/* ethernet errs */
	int uflo, oflo;
	int align_errors, fcs_errors, duplex, carrier_errors;
	int collisions,ex_collisions, tx_late_collisions;
	int defer_xmts,first_collisions,multi_collisions;
	int sqe_errors,macxmt_errors,macrcv_errors;
	int toolong_errors, runt_errors;
};


/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/

static int debug=1;

/* Your generic "puts" substitute. Prints to console */
static void routineprint(char *string){
	cmn_err(CE_WARN, "fanet: In routine %s\n", string);
}

/* if not in debug mode, this will do nothing. in theory. */
static void debugprint(char *string){
	if(debug==0) return;
	cmn_err(CE_NOTE, "%s", string);
}


/* simple way to send a oneline message of WARNING priority to console*/
static void errprint(char *string){
	cmn_err(CE_WARN, "%s",string);
}


/* This is called both by fanet_detach, and if anything goes wrong
 * with the attach routine.
 * Its always a good idea to have centralized cleanup, when things get
 * messy.
 * Be ultra-paranoid about whether something SHOULD be cleaned up, before
 * trying to actually free it.
 */
static void cleanup(struct fanet_softstate *softstate){
	int instance;
	if(softstate==NULL){
		return;
	}

	instance=ddi_get_instance(softstate->dip);

	if(softstate->gld_info!=NULL){
		gld_mac_free(softstate->gld_info);
	}
	if(softstate->fanet_regs_handle!=NULL){
		ddi_regs_map_free(&softstate->fanet_regs_handle);
	}

	mutex_destroy(&softstate->instance_mutex);
	
	ddi_soft_state_free(fanet_glob_soft_handle, instance);
}


static uint8_t ether_broadcast[7]={
	0xff,0xff,0xff,0xff,0xff,0xff,0
};

/**********************************************************************/
/*   Here are the GLD-specific entrypoints. Not to be confused with   */
/*  the solaris-generic entrypoints, which are further down           */
/**********************************************************************/

int fanet_gldm_reset(gld_mac_info_t *gldinfo){
	routineprint("gldm_reset");
	return GLD_SUCCESS;
}
int fanet_gldm_start(gld_mac_info_t *gldinfo){
	routineprint("gldm_start");
	return GLD_SUCCESS;
}
int fanet_gldm_stop(gld_mac_info_t *gldinfo){
	routineprint("gldm_stop");
	return GLD_SUCCESS;
}
int fanet_gldm_set_mac_addr(gld_mac_info_t *gldinfo){
	routineprint("gldm_set_mac_addr");
	return GLD_SUCCESS;
}
int fanet_gldm_set_multicast(gld_mac_info_t *gldinfo){
	routineprint("gldm_set_multicast");
	return GLD_SUCCESS;
}
int fanet_gldm_set_promiscuous(gld_mac_info_t *gldinfo){
	routineprint("gldm_set_promiscuous");
	return GLD_SUCCESS;
}

/* This gets called by the GLD internals when someone wants 
 * to send a packet out. Duh :-)
 * Recieving data needs to be handled (elsewhere in here) by
 * handing off the packet to gld_recv(), probably in fanet_gldm_intr
 *
 */
int fanet_gldm_send(gld_mac_info_t *gldinfo){
	routineprint("gldm_send");
	return GLD_SUCCESS;
}

/* This should actually check the device, and see if it triggered an
 * interrupt or not
 * This is called by gld_intr.
 * We should always register gld_intr as the interrupt handler, NOT
 *  gldm_intr. With the exception of high-level interrupts, which
 *  we take with fanet_handle_HIintr, then pass off to gld_intr
 *  via soft_interrupt
 *
 * according to gld(9f), if a packet has arrived,  we need to
 * hand it off to gld_recv() via STREAMS communication,
 * after releasing any locks we have
 *
 */
uint_t fanet_gldm_intr(gld_mac_info_t *gldinfo){
	routineprint("gldm_intr");
	return DDI_INTR_UNCLAIMED;
}



int fanet_gldm_get_stats(gld_mac_info_t *gldinfo){
	routineprint("gldm_get_stats");
	return GLD_SUCCESS;
}

/* End of GLD entry-points */

/**********************************************************************/
/**********************************************************************/
/* All generic solaris driver entrypoints are BELOW here              */
/* Similarly, all customized stuff is ABOVE here                      */
/**********************************************************************/
/**********************************************************************/


/*Like the fanet_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See fanet_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	routineprint("_init");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&fanet_global_mutex, "BASE Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct fanet_softstate *BASE_MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&fanet_glob_soft_handle,
				sizeof(struct fanet_softstate),
			        BASE_MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&fanet_global_mutex);
		errprint("init: modinstall !=0");
		cmn_err(CE_NOTE, "fanet: _init FAILED\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE, "fanet: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	cmn_err(CE_NOTE, "fanet: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	cmn_err(CE_NOTE, "fanet: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&fanet_glob_soft_handle);
		mutex_destroy(&fanet_global_mutex);
	}
	return (ret);
}

/************************************************************/


/*************************************************************
 *Hack for handling high-level interrupts specifically 
 * Following sun doc recommendations, we hand off to SOFTinterrupt handler
 * Which actually goes to the solaris gld_intr, and then gets
 * passed back to us in fanet_gldm_intr
 *
 *************************************************************/

static uint_t fanet_handle_HIintr(caddr_t intr_arg){
	struct fanet_softstate *softstate=(struct fanet_softstate*) intr_arg;
	
	softstate->intr_hardcount +=1;

	/* XXX in theory, should check whether this was really for us
	 * But I have no idea on how to do that rightnow
	 * So pass EVERYTHING on to softintr.
	 * Wow, this is going to be a really non-efficient driver
	 */
	mutex_enter(&softstate->HIintr_mutex);
		softstate->soft_trigger +=1;
	mutex_exit(&softstate->HIintr_mutex);
	ddi_trigger_softintr (softstate->softintr_id);

	return DDI_INTR_CLAIMED;
}

/************************************************************/


/* This is an "official" routine?!!  but not really...?*/
static int
fanet_print(dev_t dev, char *string)
{
/*
	int instance = getminor(dev);
	fanet_devstate_t *rsp = ddi_get_soft_state(fanet_state, instance);
	cmn_err(CE_WARN, "%s%d: %s\n", ddi_get_name(rsp->dip), instance, str);
*/
	cmn_err(CE_WARN, "fanet: %s\n", string);
	return (0);
}


static int	fanet_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	routineprint("getinfo");
	return DDI_FAILURE;
}




/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static	int	fanet_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	routineprint("probe");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
   	cmn_err(CE_NOTE, "probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in fanet_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  fanet_detach()
 */
static	int	fanet_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance,status;
	struct fanet_softstate *softstate;

	cmn_err(CE_NOTE, "fanet: fanet_attach called\n");
   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "fanet: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "fanet: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "fanet: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	cmn_err(CE_NOTE, "fanet: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(fanet_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "fanet: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(fanet_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "fanet: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(fanet_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}
	softstate->dip=dip;

	
	status=ddi_regs_map_setup(dip, 1,
		(caddr_t *)&softstate->fanet_regs,
		0, 8,&dev_attr, &softstate->fanet_regs_handle);


	if(status!=DDI_SUCCESS){
		if(status==DDI_REGS_ACC_CONFLICT){
			cmn_err(CE_WARN,"fanet: regs mapping conflicted with other mapping");
		} else {
			cmn_err(CE_WARN,"fanet: regs mapping failed");
		}
		ddi_soft_state_free(fanet_glob_soft_handle, instance);
		return DDI_FAILURE;
	}


	/* XXX do we really need a cookie here instead of NULL? */
	mutex_init(&softstate->instance_mutex,"instance mutex",
		   MUTEX_DRIVER, NULL);

	debugprint("Trying for gld stuff now");

	softstate->gld_info=gld_mac_alloc(dip);
	if(softstate->gld_info ==NULL){
		errprint("ERROR: gld_mac_alloc failed?");
		cleanup(softstate);
		return DDI_FAILURE;
	}
	softstate->gld_info->gldm_private=(void*)softstate;

	softstate->gld_info->gldm_reset=fanet_gldm_reset;
	softstate->gld_info->gldm_start=fanet_gldm_start;
	softstate->gld_info->gldm_stop=fanet_gldm_stop;
	softstate->gld_info->gldm_set_mac_addr=fanet_gldm_set_mac_addr;
	softstate->gld_info->gldm_set_multicast=fanet_gldm_set_multicast;
	softstate->gld_info->gldm_set_promiscuous=fanet_gldm_set_promiscuous;
	softstate->gld_info->gldm_send=fanet_gldm_send;
	softstate->gld_info->gldm_intr=fanet_gldm_intr;
	softstate->gld_info->gldm_get_stats=fanet_gldm_get_stats;
	softstate->gld_info->gldm_ioctl=NULL; /* driver-specific ioctls */

	softstate->gld_info->gldm_ident="test fanet driver";
	softstate->gld_info->gldm_type=DL_ETHER;
	softstate->gld_info->gldm_maxpkt=1500;
	softstate->gld_info->gldm_minpkt=0;
	softstate->gld_info->gldm_addrlen=6;
	softstate->gld_info->gldm_saplen=-2;

	softstate->gld_info->gldm_broadcast_addr=ether_broadcast;
	softstate->gld_info->gldm_vendor_addr="fanet_vendor_addr";
	softstate->gld_info->gldm_ppa=instance;
	softstate->gld_info->gldm_devinfo=dip;

	softstate->gld_info->gldm_cookie=softstate->intr_cookie;

	

	debugprint("Should read ether of the card now");

	routineprint("Trying to register with GLD");
	if(gld_register(dip,"fanet", softstate->gld_info)==DDI_FAILURE){

		errprint("gld_register failed");
		cleanup(softstate);
		return DDI_FAILURE;
	}
	
	routineprint("GLD registered okay");

#ifdef USE_INTRS
/* XXX will need to use this someday soon! */

	if (ddi_intr_hilevel(dip, 0) != 0){
		int numintrs;
		ddi_acc_handle_t conf_handle;

		errprint("Using HIGH-LEVEL interrupt. Ugh.");
		errprint("You should consider fiddling with your ethernet card");
		errprint("until you dont see this warning");
		softstate->use_softintr=1;
	} else {
		debugprint("fanet: using REGULAR interrupts");
	}

	if(softstate->use_softintr>0){
		ddi_get_iblock_cookie(dip,0,&softstate->HIintr_cookie);
		ddi_get_soft_iblock_cookie(dip,0,&softstate->intr_cookie);

		mutex_init(&softstate->HIintr_mutex, NULL,
                     MUTEX_DRIVER, (void*)softstate->HIintr_cookie);
		mutex_init(&softstate->intr_mutex, NULL,
                     MUTEX_DRIVER, (void*)softstate->intr_cookie);

		if(ddi_add_intr(dip,0,
				&softstate->HIintr_cookie,
				NULL,
				fanet_handle_HIintr,
			       (caddr_t)softstate)  !=DDI_SUCCESS){
			errprint("Could not add highlevel interrupt");
			cleanup(softstate);
			return DDI_FAILURE;
		}

		if (ddi_add_softintr(dip, DDI_SOFTINT_HIGH,
				     &softstate->softintr_id,
				     &softstate->intr_cookie,
				     NULL,
				     gld_intr,
			       (caddr_t)softstate)  !=DDI_SUCCESS){
			errprint("Could not add SOFT interrupt");
			cleanup(softstate);
			return DDI_FAILURE;
		}
		softstate->use_softintr +=1;

	} else { /* not using soft interrupts */
		ddi_get_iblock_cookie(dip,0,&softstate->intr_cookie);
		mutex_init(&softstate->intr_mutex, NULL,
                     MUTEX_DRIVER, (void*)softstate->intr_cookie);

		if(ddi_add_intr(dip,0,
				&softstate->intr_cookie, NULL,
				gld_intr,
			       (caddr_t)softstate)  !=DDI_SUCCESS){
			cmn_err(CE_WARN,"Could not add interrupt");
			cleanup(softstate);
			return DDI_FAILURE;
		}
	}

	
#endif


	/* The GLD framework takes care of creating the
	 * minor node. AND the link in /dev. Cool!
	 */

   
	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}


static	int	fanet_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct fanet_softstate *softstate;

   	instance=ddi_get_instance(dip);
   	cmn_err(CE_NOTE, "fanet: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(fanet_glob_soft_handle,instance);

	if(gld_unregister(softstate->gld_info)==DDI_FAILURE){
		errprint("detach failed: could not gld_unregister");
		return DDI_FAILURE;
	}

	cleanup(softstate);

	ddi_remove_minor_node(dip,NULL);
	
	return (DDI_SUCCESS);
		     
}

int fanet_open(dev_t *devp, int flag, int otyp,cred_t *cred_p){
	routineprint("open");
	return DDI_SUCCESS;
}
int fanet_close(dev_t dev, int flag, int otyp,cred_t *cred_p){
	routineprint("open");
	cmn_err(CE_NOTE, "close: cred_p is %x",cred_p);
	return DDI_SUCCESS;
}
